/*
  Copyright 2021 Kakao Corp.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
 */
package com.kakao.sdk.partner.friend.view

import android.app.Activity
import android.content.Intent
import android.content.res.Configuration
import android.os.Bundle
import android.os.Parcelable
import android.os.ResultReceiver
import android.view.Gravity
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.LinearLayout
import android.widget.Toast
import androidx.constraintlayout.widget.ConstraintSet
import androidx.fragment.app.Fragment
import androidx.lifecycle.Observer
import androidx.lifecycle.ViewModelProvider
import androidx.recyclerview.widget.RecyclerView
import androidx.recyclerview.widget.SimpleItemAnimator
import com.kakao.sdk.common.model.ClientError
import com.kakao.sdk.common.model.ClientErrorCause
import com.kakao.sdk.common.model.KakaoSdkError
import com.kakao.sdk.common.util.SdkLog
import com.kakao.sdk.partner.friend.Constants
import com.kakao.sdk.partner.friend.adapter.ChatsListAdapter
import com.kakao.sdk.partner.friend.model.DisableSelectReason
import com.kakao.sdk.partner.friend.model.PickerChatRequestParams
import com.kakao.sdk.partner.friend.model.PickerChatSelectionType
import com.kakao.sdk.partner.friend.model.PickerFriendRequestParams
import com.kakao.sdk.partner.friend.util.KeyboardController
import com.kakao.sdk.partner.friend.util.LayoutUtils
import com.kakao.sdk.partner.friend.viewmodel.*
import com.kakao.sdk.v2.partner.friend.R
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkChatsListBinding
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkFragmentChatPickerBinding
import com.kakao.sdk.v2.partner.friend.databinding.KakaoSdkTitleBarBinding
import kotlin.math.max

/**
 * @suppress
 */
internal class ChatPickerFragment : Fragment() {
    private var _binding: KakaoSdkFragmentChatPickerBinding? = null
    private val binding get() = _binding!!

    private var title = ""
    private var isPopup = false
    private var isTab = false
    private var selectionType = PickerChatSelectionType.CHAT
    private var enableSearch = true
    private var maxPickableCount = 1
    private var minPickableCount = 1
    private var disableSelectReasons: List<DisableSelectReason>? = null

    private var titleBar: KakaoSdkTitleBarBinding? = null
    private var searchBar: SearchBar? = null
    private var chatsRv: KakaoSdkChatsListBinding? = null

    private val chatListAdapter: ChatsListAdapter by lazy {
        ChatsListAdapter(
            isPopup, checkCallback = viewModel::addCheckedChat
        ) {
            viewModel.removeUncheckedChat()
        }
    }

    private lateinit var viewModel: ChatViewModel
    private var tabViewModel: TabViewModel? = null

    override fun onCreateView(
        inflater: LayoutInflater, container: ViewGroup?,
        savedInstanceState: Bundle?
    ): View {
        _binding = KakaoSdkFragmentChatPickerBinding.inflate(layoutInflater, container, false)
        return binding.root
    }

    override fun onViewCreated(view: View, savedInstanceState: Bundle?) {
        super.onViewCreated(view, savedInstanceState)
        createViewModel(requireActivity().intent)
        setProperty()
        createLayout()

        initializeView()
        observeData()
    }

    private fun sendError(exception: KakaoSdkError) {
        viewModel.resultReceiver.send(
            Activity.RESULT_CANCELED,
            Bundle().apply { putSerializable(Constants.KEY_EXCEPTION, exception) }
        )
        requireActivity().finish()
    }

    private fun actionDoneButton() {
        if (selectionType == PickerChatSelectionType.CHAT_MEMBER) {
            viewModel.loadChatMember { error ->
                if (error != null) {
                    Toast.makeText(
                        requireContext(),
                        getString(R.string.user_scope_error_message),
                        Toast.LENGTH_LONG
                    ).show()
                } else {
                    // 친구 피커를 사용하기 위해 PickerRequestParams을 생성해서 넘겨줘야함
                    sendParamToActivity()
                    navigateToFriendPicker()
                }
            }
        } else {
            viewModel.sendResult()
            requireActivity().finish()
        }
    }

    private fun navigateToFriendPicker() {
        requireActivity().supportFragmentManager.beginTransaction()
            .replace(R.id.fragment_container, FriendPickerFragment())
            .addToBackStack(null).commit()
    }

    private fun sendParamToActivity() {
        requireActivity().intent =
            Intent().putExtra(Constants.KEY_BUNDLE, Bundle().apply {
                putBoolean(Constants.KEY_IS_SINGLE, false)
                putBoolean(Constants.KEY_IS_POPUP, isPopup)
                putBoolean(Constants.KEY_IS_CHAT_MEMBER_PICKER, true)
                putBoolean(Constants.KEY_IS_TAB, false)
                putParcelable(Constants.KEY_RESULT_RECEIVER, viewModel.resultReceiver)
                putParcelable(
                    Constants.KEY_FRIEND_PICKER_PARAM,
                    // TODO: 스펙에 맞춰 멀티 피커 파라미터 설정해야함. 인터페이스 구현 후 수정 예정
                    PickerFriendRequestParams(
                        showMyProfile = false,
                        showFavorite = false,
                        enableIndex = false,
                        enableSearch = enableSearch,
                        maxPickableCount = maxPickableCount,
                        minPickableCount = minPickableCount,
                        disableSelectReasons = disableSelectReasons,
                        displayAllProfile = viewModel.displayAllProfile
                    )
                )
                putParcelable(Constants.KEY_CHAT_PICKER_PARAM, viewModel.param)
            })
    }

    private fun createViewModel(intent: Intent) {
        try {
            intent.extras?.getBundle(Constants.KEY_BUNDLE)?.run {
                val resultReceiver =
                    getParcelable<ResultReceiver>(Constants.KEY_RESULT_RECEIVER) as ResultReceiver
                val param = getParcelable<Parcelable>(Constants.KEY_CHAT_PICKER_PARAM)!!

                val isPopup = getBoolean(Constants.KEY_IS_POPUP)
                val isTab = getBoolean(Constants.KEY_IS_TAB)

                viewModel = ViewModelProvider(
                    this@ChatPickerFragment,
                    ChatViewModelFactory(
                        isPopup,
                        isTab,
                        resultReceiver,
                        param as PickerChatRequestParams
                    )
                ).get(ChatViewModel::class.java)

                if (isTab) {
                    tabViewModel = ViewModelProvider(
                        requireActivity(),
                        ViewModelProvider.NewInstanceFactory()
                    ).get(TabViewModel::class.java)
                }
            }
        } catch (e: Throwable) {
            SdkLog.e(e)
            sendError(ClientError(ClientErrorCause.Unknown).apply { initCause(e) })
        }
    }

    private fun createLayout() {
        binding.mainLayout.apply {
            if (!isTab) {
                titleBar = KakaoSdkTitleBarBinding.inflate(layoutInflater, this, false).apply {
                    title.text = this@ChatPickerFragment.title
                    pickedCountTv.visibility = View.INVISIBLE
                    addView(root)
                    doneBtn.text =
                        if (selectionType == PickerChatSelectionType.CHAT) getString(R.string.ok) else getString(R.string.next)
                }
            }

            if (enableSearch) {

                searchBar = SearchBar(requireContext())
                searchBar?.let {
                    it.setPlaceholderText(getString(R.string.search_chat_hint))
                    addView(it)
                    it.setSearchAction(viewModel::searchChats)
                }
            }
            chatsRv = KakaoSdkChatsListBinding.inflate(layoutInflater, this, false).apply {
                addView(root, LinearLayout.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, 0, 1f))
            }
        }

        if (isPopup) {
            setPopupLayout()
        }
    }

    private fun setPopupLayout() {
        titleBar?.apply {
            val marginTop = if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                resources.getDimensionPixelSize(R.dimen.popup_title_bar_padding_top)
            } else {
                resources.getDimensionPixelSize(R.dimen.popup_title_bar_padding_top_landscape)
            }
            root.setPadding(0, marginTop, 0, 0)
            title.apply {
                gravity = Gravity.CENTER

                // 뷰가 그려지기 전에 width를 계산
                backBtn.measure(View.MeasureSpec.UNSPECIFIED, View.MeasureSpec.UNSPECIFIED)
                doneBtnLayout.measure(View.MeasureSpec.UNSPECIFIED, View.MeasureSpec.UNSPECIFIED)

                // 타이틀 좌, 우 뷰 중에서 더 넓은 뷰 넓이에 마진값 적용
                val marginHorizontal = max(
                    backBtn.measuredWidth,
                    doneBtnLayout.measuredWidth
                ) + resources.getDimensionPixelSize(R.dimen.popup_title_bar_text_margin_start)

                ConstraintSet().apply {
                    clone(root)
                    connect(
                        id,
                        ConstraintSet.START,
                        ConstraintSet.PARENT_ID,
                        ConstraintSet.START, marginHorizontal
                    )
                    connect(
                        id,
                        ConstraintSet.END,
                        ConstraintSet.PARENT_ID,
                        ConstraintSet.END, marginHorizontal
                    )
                }.applyTo(root)
            }
        }
        chatsRv?.chatsList?.apply {
            val paddingBottom = if (LayoutUtils.orientation == Configuration.ORIENTATION_PORTRAIT) {
                resources.getDimensionPixelSize(R.dimen.popup_friends_list_padding_bottom)
            } else {
                resources.getDimensionPixelSize(R.dimen.popup_friends_list_padding_bottom_landscape)
            }
            setPadding(0, 0, 0, paddingBottom)
        }
    }

    private fun initializeView() {
        chatsRv?.chatsList?.apply {
            adapter = chatListAdapter
            itemAnimator.apply {
                if (this is SimpleItemAnimator) {
                    supportsChangeAnimations = false
                }
            }
            addOnScrollListener(object : RecyclerView.OnScrollListener() {
                override fun onScrollStateChanged(recyclerView: RecyclerView, newState: Int) {
                    super.onScrollStateChanged(recyclerView, newState)
                    searchBar?.let {
                        it.editText.clearFocus()
                        KeyboardController.hideKeyboard(requireContext(), it.editText)
                    }
                }
            })
        }

        titleBar?.apply {
            backBtn.setOnClickListener {
                sendError(ClientError(ClientErrorCause.Cancelled, "cancelled."))
            }
            doneBtn.setOnClickListener {
                actionDoneButton()
            }
        }
    }

    private fun observeData() {
        viewModel.run {
            pickerList.observe(viewLifecycleOwner, Observer {
                chatListAdapter.setList(it)
            })

            searchStatus.observe(viewLifecycleOwner, Observer {
                if (it == SearchStatus.SEARCHING) {
                    chatsRv?.apply {
                        chatsList.scrollToPosition(0)

                    }
                } else if (it == SearchStatus.DONE) {
                    chatsRv?.apply {
                        chatsList.scrollToPosition(0)
                    }
                }
            })

            pickedChat.observe(viewLifecycleOwner, Observer { chat ->
                if (chat == null) {
                    chatListAdapter.uncheckChat()
                    titleBar?.doneBtn?.isEnabled = false
                } else {
                    chatListAdapter.checkChat(chat)
                    titleBar?.doneBtn?.isEnabled = true
                }
                chatListAdapter.setPickedChat(chat)
                tabViewModel?.setChatPicked(chat != null)
            })

            isError.observe(viewLifecycleOwner, Observer {
                if (it) {
                    titleBar?.doneBtn?.visibility = View.GONE
                    searchBar?.visibility = View.GONE
                    binding.emptyChatView.setErrorMessage(getString(R.string.empty_chat_message))

                    if (!isTab) {
                        binding.emptyChatView.visibility = View.VISIBLE
                    } else {
                        tabViewModel?.setEmptyChatError(it)
                    }
                } else {
                    titleBar?.doneBtn?.visibility = View.VISIBLE
                    searchBar?.visibility = View.VISIBLE

                    if (!isTab) {
                        binding.emptyChatView.visibility = View.GONE
                    } else {
                        tabViewModel?.setEmptyChatError(it)
                    }
                }
            })
        }
        tabViewModel?.chatStatus?.observe(viewLifecycleOwner, Observer {
            when (it) {
                TabStatus.DONE -> actionDoneButton()
                TabStatus.CANCELED -> {
                    sendError(ClientError(ClientErrorCause.Cancelled, "cancelled."))
                }
                TabStatus.CLEARED -> viewModel.clearPickedChat()
                else -> {
                    // Do Nothing
                }
            }
        })
    }

    private fun setProperty() {
        isPopup = viewModel.isPopup
        isTab = viewModel.isTab

        viewModel.param.also {
            title = it.title ?: getString(R.string.select_chat)
            selectionType = it.selectionType
            enableSearch = it.enableSearch ?: Constants.DEFAULT_ENABLE_SEARCH
            minPickableCount = it.minPickableCount ?: Constants.DEFAULT_MIN_PICKABLE_COUNT
            maxPickableCount = it.maxPickableCount ?: Constants.DEFAULT_MAX_PICKABLE_COUNT
            disableSelectReasons = it.disableSelectReasons
        }
    }

    override fun onDestroyView() {
        super.onDestroyView()
        _binding = null
    }
}